package jemu.system.cpc;

import java.applet.Applet;
import java.awt.*;
import java.awt.event.KeyEvent;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import jemu.core.Util;
import jemu.core.cpu.Processor;
import jemu.core.device.Computer;
import jemu.core.device.DeviceMapping;
import jemu.ui.Switches;
import jemu.core.device.crtc.Basic6845;
import jemu.core.device.floppy.Drive;
import jemu.core.device.floppy.UPD765A;
import jemu.core.device.io.PPI8255;
import jemu.core.device.memory.Memory;
import jemu.core.device.sound.AY_3_8910;
import jemu.ui.Display;
import jemu.util.diss.Disassembler;
import jemu.util.diss.DissZ80;
import jemu.ui.Autotype;
import jemu.ui.CPCPrinter;

/**
 * Title:        JEMU
 * Description:  The Java Emulation Platform
 * Copyright:    Copyright (c) 2002
 * Company:
 * @author
 * @version 1.0
 */

public class CPC extends Computer {
    
    public CPCDiscImage dskImage;
    public CPCDiscImage dskImageA;
    public CPCDiscImage dskImageB;
    public CPCDiscImage dskImageC;
    public CPCDiscImage dskImageD;

  public int joyreader;
  protected boolean turbo = false;
  public int autotyper = 0;
  public int readkey  = 0;
  public static boolean blockKeyboard;
  public static int getfromautotype = 0;

  int[] eventArray;
  boolean[] shifter;
  public static boolean shift;

  // Conversion of CPC BDIR & BC1 values 0..3 to PSG values
  protected static final int[] PSG_VALUES = new int[] {
    AY_3_8910.BC2_MASK,
    AY_3_8910.BC2_MASK | AY_3_8910.BC1_MASK,
    AY_3_8910.BC2_MASK | AY_3_8910.BDIR_MASK,
    AY_3_8910.BC2_MASK | AY_3_8910.BDIR_MASK | AY_3_8910.BC1_MASK
  };

  // Port mappings
  protected static final int   PSG_PORT_A        = -1;
  protected static final int   PPI_PORT_B        = -2;
  protected static final int   PPI_PORT_C        = -3;
  
  public static int autotype = 4;

  protected static final int   CYCLES_PER_SECOND = 1000000;
  protected static final int   AUDIO_TEST        = 0x40000000;

  protected String             lowerROM          = null;
  protected Hashtable          upperROMs         = new Hashtable();

  protected Z80                z80               = new Z80(CYCLES_PER_SECOND);
  protected CPCMemory          memory            = null;                                                      // new
  // CPCMemory(CPCMemory.TYPE_512K);
  protected Basic6845          crtc              = (Basic6845) addDevice(new Basic6845());
  protected GateArray          gateArray         = (GateArray) addDevice(new GateArray(this));
  protected PPI8255            ppi               = (PPI8255) addDevice(new PPI8255());
  protected AY_3_8910          psg               = (AY_3_8910) addDevice(new AY_3_8910());
  protected UPD765A            fdc               = (UPD765A) addDevice(new UPD765A(4));  // 4 cycles per call to cycle()
  protected Disassembler       disassembler      = new DissZ80();
  protected int                audioAdd          = psg.getSoundPlayer().getClockAdder(AUDIO_TEST,
                                                                                           CYCLES_PER_SECOND);
  protected int                audioCount        = 0;
  protected Drive[]            floppies          = new Drive[4];
  protected int                currentDrive      = 0;

  protected KeyboardB          keyboardb         = new KeyboardB();
  protected KeyboardA          keyboarda         = new KeyboardA();
  protected KeyboardN          keyboardn         = new KeyboardN();

  public CPC(Applet applet, String name) {
    super(applet, name);
    initCPCType(name);
    z80.setMemoryDevice(memory);
    z80.addOutputDeviceMapping(new DeviceMapping(memory, 0x2000, 0x0000)); // ROM Select
    z80.setInterruptDevice(gateArray);
    z80.addOutputDeviceMapping(new DeviceMapping(gateArray, 0xc000, 0x4000)); // All GA functions
    z80.setCycleDevice(this);
    crtc.setRegisterSelectMask(0x0100, 0x0000);
    crtc.setCRTCListener(gateArray);
    z80.addOutputDeviceMapping(new DeviceMapping(crtc, 0x4000, 0x0000));
    ppi.setPortMasks(0x0100, 0x0100, 0x0200, 0x0200);
    ppi.setReadDevice(PPI8255.PORT_B, this, PPI_PORT_B);
    ppi.setWriteDevice(PPI8255.PORT_C, this, PPI_PORT_C);
    ppi.setReadDevice(PPI8255.PORT_A, psg, 0);
    ppi.setWriteDevice(PPI8255.PORT_A, psg, 0);
    psg.setReadDevice(AY_3_8910.PORT_A, this, PSG_PORT_A);
    psg.setClockSpeed(CYCLES_PER_SECOND);
    z80.addOutputDeviceMapping(new DeviceMapping(ppi, 0x0800, 0x0000));
    z80.addInputDeviceMapping(new DeviceMapping(ppi, 0x0800, 0x0000));
    z80.addOutputDeviceMapping(new DeviceMapping(fdc, 0x0580, 0x0100));
    z80.addInputDeviceMapping(new DeviceMapping(fdc, 0x0580, 0x0100));
    z80.addOutputDeviceMapping(new DeviceMapping(this, 0x581, 0x0000));
    for (int i = 0; i < 4; i++)
      fdc.setDrive(i, floppies[i] = new Drive(i == 0 ? 1 : 2));
    setBasePath("cpc");
  }
  
  private void initCPCType(String name) {
            
    // check for MAXAM assembler
    if (Switches.Maxam)
      upperROMs.put("14", "MAXAM.zip");    
    
    // Set memorytype for all models
      memory = new CPCMemory(CPCMemory.TYPE_512K);
      if (Switches.Memory.equals("TYPE_512K"))
        memory = new CPCMemory(CPCMemory.TYPE_512K);
      if (Switches.Memory.equals("TYPE_256K"))
        memory = new CPCMemory(CPCMemory.TYPE_256K);
      if (Switches.Memory.equals("TYPE_128K"))
        memory = new CPCMemory(CPCMemory.TYPE_128K);
      if (Switches.Memory.equals("TYPE_64K"))
        memory = new CPCMemory(CPCMemory.TYPE_64K);
      if (Switches.Memory.equals("TYPE_SILICON_DISC"))
        memory = new CPCMemory(CPCMemory.TYPE_SILICON_DISC);
    //
      
      if ("CPC464".equals(name)) {
      lowerROM = "OS464.zip";
      upperROMs.put("0", "BASIC1-0.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPC664".equals(name)) {
      lowerROM = "OS664.zip";
      upperROMs.put("0", "BASIC664.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPC464PARA".equals(name)) {
      lowerROM = "OS464.zip";
      upperROMs.put("0", "BASIC1-0.zip");
      upperROMs.put("7", "PARADOS.zip");
    } else if ("CPC6128".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("ST128".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "STK1.zip");
      upperROMs.put("2", "STK2.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("ART".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "ART0.zip");
      upperROMs.put("2", "ART1.zip");
      upperROMs.put("3", "ART2.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("MEGA".equals(name)) {
      lowerROM = "MOS6128.ZIP";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "STK1.zip");
      upperROMs.put("2", "STK2.zip");
      upperROMs.put("3", "ART0.zip");
      upperROMs.put("4", "ART1.zip");
      upperROMs.put("5", "ART2.zip");
      upperROMs.put("6", "locksmit.zip");
      upperROMs.put("7", "AMSDOS.zip");
      upperROMs.put("8", "sym-romA.zip");
      upperROMs.put("9", "sym-romB.zip");
      upperROMs.put("10", "sym-romC.zip");
      upperROMs.put("11", "sym-romD.zip");
      upperROMs.put("12", "ROMPKP1.zip");
      upperROMs.put("13", "ROMPKP2.zip");
    } else if ("PARADOS".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "PARADOS.zip");
    } else if ("ROMPACK".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "ROMPKP1.zip");
      upperROMs.put("2", "ROMPKP2.zip");
      upperROMs.put("7", "PARADOS.zip");
    } else if ("ROMSYM".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "sym-romA.zip");
      upperROMs.put("2", "sym-romB.zip");
      upperROMs.put("3", "sym-romC.zip");
      upperROMs.put("4", "sym-romD.zip");
      upperROMs.put("5", "ROMPKP1.zip");
      upperROMs.put("6", "ROMPKP2.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPC6128fr".equals(name)) {
      lowerROM = "OSFR.zip";
      upperROMs.put("0", "BASICFR.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPC6128es".equals(name)) {
      lowerROM = "OSES.zip";
      upperROMs.put("0", "BASICES.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("SymbOS".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
      upperROMs.put("1", "sym-romA.zip");
      upperROMs.put("2", "sym-romB.zip");
      upperROMs.put("3", "sym-romC.zip");
      upperROMs.put("4", "sym-romD.zip");
      } else if ("FutureOS".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
      upperROMs.put("10", "T-OG-E-A.zip");
      upperROMs.put("11", "T-OG-E-B.zip");
      upperROMs.put("12", "T-OG-E-C.zip");
      upperROMs.put("13", "T-OG-E-D.zip");
    } else if ("KCcomp".equals(name)) {
      lowerROM = "KCCOS.zip";
      upperROMs.put("0", "KCCBAS.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("COMCPC".equals(name)) {
      lowerROM = "-OS.zip";
      upperROMs.put("0", "-BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("JEMCPC".equals(name)) {
      lowerROM = "jemuOS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPCPLUS".equals(name)) {
      lowerROM = "os-plus.zip";
      upperROMs.put("0", "basic-plus.zip");
      upperROMs.put("7", "romplus.zip");
    } else if ("RAMSETTER".equals(name)) {
      lowerROM = "OS6128.zip";
    }
    // Manual rom settings
    // in progress
      
    
    else if ("MANUAL".equals(name)) {
      if (!Switches.LOWER.equals("none"))    lowerROM = Switches.LOWER;
      if (!Switches.UPPER_0.equals("none"))  upperROMs.put("0", Switches.UPPER_0);
      if (!Switches.UPPER_1.equals("none"))  upperROMs.put("1", Switches.UPPER_1);
      if (!Switches.UPPER_2.equals("none"))  upperROMs.put("2", Switches.UPPER_2);
      if (!Switches.UPPER_3.equals("none"))  upperROMs.put("3", Switches.UPPER_3);
      if (!Switches.UPPER_4.equals("none"))  upperROMs.put("4", Switches.UPPER_4);
      if (!Switches.UPPER_5.equals("none"))  upperROMs.put("5", Switches.UPPER_5);
      if (!Switches.UPPER_6.equals("none"))  upperROMs.put("6", Switches.UPPER_6);
      if (!Switches.UPPER_7.equals("none"))  upperROMs.put("7", Switches.UPPER_7);
      if (!Switches.UPPER_8.equals("none"))  upperROMs.put("8", Switches.UPPER_8);
      if (!Switches.UPPER_9.equals("none"))  upperROMs.put("9", Switches.UPPER_9);
      if (!Switches.UPPER_A.equals("none"))  upperROMs.put("10", Switches.UPPER_A);
      if (!Switches.UPPER_B.equals("none"))  upperROMs.put("11", Switches.UPPER_B);
      if (!Switches.UPPER_C.equals("none"))  upperROMs.put("12", Switches.UPPER_C);
      if (!Switches.UPPER_D.equals("none"))  upperROMs.put("13", Switches.UPPER_D);
      if (!Switches.UPPER_E.equals("none"))  upperROMs.put("14", Switches.UPPER_E);
      if (!Switches.UPPER_F.equals("none"))  upperROMs.put("15", Switches.UPPER_F);
    }
    Switches.computersys = 1;
  }

  
  public void initialise() {

    String romNum;
    String romName;
    memory.setLowerROM(getFile(romPath + lowerROM));
    Set roms = upperROMs.keySet();
    Iterator romIt = roms.iterator();
    while (romIt.hasNext()) {
      romNum = (String) romIt.next();
      romName = (String) upperROMs.get(romNum);
      memory.setUpperROM((new Byte(romNum)).byteValue(), getFile(romPath + romName));
    }
    upperROMs.clear();
    gateArray.setMemory(memory.getMemory());
    super.initialise();
    psg.getSoundPlayer().play();
    fdc.initialise();
    for (int i = 0; i < 4; i++)
      fdc.setDrive(i, null);
  }

  
  public void reset() {
          if (turbo)
              Switches.turbo = 3;
          else 
              Switches.turbo = 1;
          
   Display.autotype=0;
      try {
          keyboarda.keyReleased(eventArray[readkey]);
      }
      catch (Exception e) {}
    blockKeyboard = false;
    autotyper = 0;
    super.reset();
    fdc.reset();
    memory.reset();
    psg.getSoundPlayer().resync();
    fdc.initialise();
  }

  
  public void eject() {
    fdc.setDrive(getCurrentDrive(), null);
    int drive = getCurrentDrive();
    if (drive == 0)
        dskImageA = null;
    if (drive == 1)
        dskImageB = null;
    if (drive == 2)
        dskImageC = null;
    if (drive == 3)
        dskImageD = null;
    if (Switches.floppysound)
        UPD765A.ejectSound.play();
    {
      if (getCurrentDrive() == 0)
          Switches.loaddrivea = "Drive is empty.";
      if (getCurrentDrive() == 1)
          Switches.loaddriveb = "Drive is empty.";
      if (getCurrentDrive() == 2)
          Switches.loaddrivec = "Drive is empty.";
      if (getCurrentDrive() == 3)
          Switches.loaddrived = "Drive is empty.";
    }
  }

  
  public void dispose() {
    super.dispose();
    psg.getSoundPlayer().dispose();
  }
  
  public void cycle() {
    fdc.cycle();
    gateArray.cycle();
    if ((audioCount += audioAdd / Switches.turbo) >= AUDIO_TEST) {
      psg.writeAudio();
      audioCount -= AUDIO_TEST;
    }
  }

  public void setDisplay(Display value) {
    super.setDisplay(value);
    gateArray.setDisplay(value);
  }

  public void setFrameSkip(int value) {
    super.setFrameSkip(value);
    gateArray.setRendering(value == 0);
  }

  public void vSync() {
    if (frameSkip == 0)
      display.updateImage(true);
    syncProcessor(psg.getSoundPlayer());
    typeAuto();
    joyReader();
    
      if (Switches.dskcheck == true){
          Switches.dskcheck = false;
          SaveDSK();
      }
  }
  
  public void SaveDSK(){
      dskImage = null;
      int drive = getCurrentDrive();
      System.out.println("Saving image from drive DF" + drive);
      if (drive == 0){
          if (dskImageA != null)
                dskImage=dskImageA;
      }
      if (drive == 1){
          if (dskImageB != null)
                dskImage=dskImageB;
      }
      if (drive == 2){
          if (dskImageC != null)
                dskImage=dskImageC;
      }
      if (drive == 3){
          if (dskImageD != null)
                dskImage=dskImageD;
      }
      if (dskImage != null){
        floppies[drive].setSides(dskImage.getNumberOfSides());
        int heads = dskImage.getNumberOfSides() == 1 ? Drive.HEAD_0 : Drive.BOTH_HEADS;
        fdc.setDrive(drive, floppies[drive]);
        fdc.getDrive(drive).setDisc(heads, dskImage);
        dskImage.saveImage();
      } 
      else
          System.out.println("Failed to save... Probably drive empty?");
      setCurrentDrive(0);
  }
  
  public void joyReader(){
       if (!blockKeyboard){
      if (Switches.MouseJoy){
      if (Switches.directxR.equals("Right"))
          keyboarda.keyPressed(KeyEvent.VK_NUMPAD6);
      if (Switches.directxL.equals("Left"))
          keyboarda.keyPressed(KeyEvent.VK_NUMPAD4);
      if (Switches.directyU.equals("Up"))
          keyboarda.keyPressed(KeyEvent.VK_NUMPAD8);
      if (Switches.directyD.equals("Down"))
          keyboarda.keyPressed(KeyEvent.VK_NUMPAD2);
              Switches.directL++;
              if (Switches.directL >= 4){
                  keyboarda.keyReleased(KeyEvent.VK_NUMPAD4);
              }
              Switches.directR++;
              if (Switches.directR >= 4){
                  keyboarda.keyReleased(KeyEvent.VK_NUMPAD6);
              }
              Switches.directU++;
              if (Switches.directU >= 4){
                  keyboarda.keyReleased(KeyEvent.VK_NUMPAD8);
              }
              Switches.directD++;
              if (Switches.directD >= 4){
                  keyboarda.keyReleased(KeyEvent.VK_NUMPAD2);
              }
      }    
       }
  }

  public void typeAuto(){
    if (getfromautotype != 0) {
      getfromautotype = 0;
      AutoTyper(Autotype.autotext);
    }
    if (autotyper != 0) {
        Switches.turbo=3;
      Display.autotype=2;
      autotyper++;
      if (autotyper == 3) {
        try {
         // System.out.println("KeyRelease: " + readkey);
          keyboarda.keyReleased(eventArray[readkey]);
          if (shifter[readkey + 1])
            keyboarda.keyPressed(KeyEvent.VK_SHIFT);
          else
            keyboarda.keyReleased(KeyEvent.VK_SHIFT);
        }  catch (Exception e) {
          keyboarda.keyReleased(KeyEvent.VK_SHIFT);
          autotyper = 0;
          Display.autotype=0;
          if (turbo)
              Switches.turbo = 3;
          else 
              Switches.turbo = 1;
          blockKeyboard = false;
        }
      }
        
      if (autotyper == 5) {
        try {
          //System.out.println("KeyPress: " + readkey);
          readkey++;
          keyboarda.keyPressed(eventArray[readkey]);
          autotyper = 1;
        } catch (Exception e) {
          keyboarda.keyReleased(KeyEvent.VK_SHIFT);
          autotyper = 0;
          Display.autotype=0;
          if (turbo)
              Switches.turbo = 3;
          else 
              Switches.turbo = 1;
          blockKeyboard = false;
        } 
      }
    }
  }

  public Memory getMemory() {
    return memory;
  }

  public Processor getProcessor() {
    return z80;
  }

  public Dimension getDisplaySize(boolean large) {
    return gateArray.getDisplaySize(large);
  }
  
  public Dimension getDisplayScale(boolean large) {
    return large ? Display.SCALE_1x2 : Display.SCALE_1;
  }
  
  public void setLarge(boolean value) {
    gateArray.setHalfSize(!value);
  }

  public Disassembler getDisassembler() {
    return disassembler;
  }
  
  protected Keyboard getKeyboard() {
    if (Switches.joystick == 0)
      return keyboardb;
    else if (Switches.notebook)
      return keyboardn;
    else
      return keyboarda;
  }

  public int readPort(int port) {
int result;
    switch(port) {
      case PPI_PORT_B:
          if (Switches.Printer)
        result = 0x1e | (crtc.isVSync() ? 0x01 : 0);
          else              
        result = 0x5e | (crtc.isVSync() ? 0x01 : 0);
              
        break;
        
      case PSG_PORT_A: {
        if (blockKeyboard) {
          result = keyboarda.readSelectedRow();
          break;
        }
        else {
          result = getKeyboard().readSelectedRow();
          break;
        }
      }
      
      default:
        throw new RuntimeException("Unexpected Port Read: " + Util.hex((short)port));
    }
    return result;
  }

  public void writePort(int port, int value) {
    switch(port) {

      case PPI_PORT_C:
        psg.setBDIR_BC2_BC1(PSG_VALUES[value >> 6],ppi.readOutput(PPI8255.PORT_A));
      
        if (blockKeyboard)
          keyboarda.setSelectedRow(value & 0x0f);
        else
          getKeyboard().setSelectedRow(value & 0x0f);
        break;

      default:
        if ((port & 0x0581) == 0) { // FDC Motor control
          System.out.println("Floppy-motor " + ((value & 0x01) == 0 ? "off" : "on"));
          if ((value & 0x01) == 0) {
            if (Switches.floppysound)
              UPD765A.motorSound.play();
            Display.ledOn = 0;
          }
          else {
            if (Switches.floppysound){
              UPD765A.motorSound.loop();
            Display.ledOn = 1;}
          }
        } else {
          throw new RuntimeException("Unexpected Port Write: " + Util.hex((short) port) + " with "
            + Util.hex((byte) value));
      }
    }
  }

  public void keyPressed(KeyEvent e) {
    if (!blockKeyboard)
      getKeyboard().keyPressed(e.getKeyCode());
   // System.out.println(e.getKeyCode());
  }

  public void keyReleased(KeyEvent e) {
    if (!blockKeyboard)
      getKeyboard().keyReleased(e.getKeyCode());
  }
  
  public void AtKey() {
    AutoTyper("@");
  }
  
  public void bootDisk() {
      Display.bootgames = 150;
    AutoTyper("mode 1:pen 0\n|A:CAT:FOR k=1 TO 8:n$=n$+chr$(peek(HIMEM-&7FF+k)):NEXT:" +
            "n$=n$+\".\":FOR k=1 TO 3:n$=n$+chr$(peek(HIMEM-&7FF+k+8)):NEXT:PEN 1:RUN n$\n");
  }  
  public void bootDiskb() {
      Display.bootgames = 150;
    AutoTyper("mode 1:pen 0\n|B:CAT:FOR k=1 TO 8:n$=n$+chr$(peek(HIMEM-&7FF+k)):NEXT:" +
            "n$=n$+\".\":FOR k=1 TO 3:n$=n$+chr$(peek(HIMEM-&7FF+k+8)):NEXT:PEN 1:RUN n$\n");
  }  
  
  public void bootCPM() {
    AutoTyper("|CPM\n");
  }
  
  public void typeRun() {
    AutoTyper("RUN\"");      
  }
  
  public void typeCat() {
    AutoTyper("CAT\n");      
  }
  
  public void AutoType() {
    AutoTyper(Switches.loadauto);      
  }  
  
  public void MouseFire1(){
       if (!blockKeyboard)
      keyboarda.keyPressed(KeyEvent.VK_NUMPAD5);
  }
  public void MouseFire2(){
       if (!blockKeyboard)
      keyboarda.keyPressed(KeyEvent.VK_NUMPAD0);
  }
  public void MouseReleaseFire1(){
       if (!blockKeyboard)
      keyboarda.keyReleased(KeyEvent.VK_NUMPAD5);
  }
  public void MouseReleaseFire2(){
       if (!blockKeyboard)
      keyboarda.keyReleased(KeyEvent.VK_NUMPAD0);
      
  }
  public void AutoTyper(String cmd) {
          if (Switches.turbo == 3)
        turbo = true;
    else
        turbo = false;
    blockKeyboard = true;
    readkey = 0;
    eventArray = new int[cmd.length()];
    shifter = new boolean[cmd.length()];
    for(int i = 0; i < cmd.length(); i++) {
      eventArray[i] = convertCharToVK(cmd.charAt(i));
      shifter[i] = shift;
    }
    if (shifter[0])
      keyboarda.keyPressed(KeyEvent.VK_SHIFT);
    keyboarda.keyPressed(eventArray[0]);
    autotyper = 1;
  }
  
  /**
   * This should really be done using CPC key codes and overriding the value returned in
   * readPort for the PSG port read.
   */
  protected static final int[] ASC_TO_KEY = {
    KeyEvent.VK_SPACE, KeyEvent.VK_1, KeyEvent.VK_2, KeyEvent.VK_3,                       // SPACE !"#   FTTT
    KeyEvent.VK_4, KeyEvent.VK_5, KeyEvent.VK_6, KeyEvent.VK_7,                           // $%&'        TTTT
    KeyEvent.VK_8, KeyEvent.VK_9, KeyEvent.VK_SEMICOLON, KeyEvent.VK_QUOTE,               // ()*+        TTTT
    KeyEvent.VK_COMMA, KeyEvent.VK_MINUS, KeyEvent.VK_PERIOD, KeyEvent.VK_SLASH,          // ,-./        FFFF
    KeyEvent.VK_0, KeyEvent.VK_1, KeyEvent.VK_2, KeyEvent.VK_3,                           // 0123        FFFF
    KeyEvent.VK_4, KeyEvent.VK_5, KeyEvent.VK_6, KeyEvent.VK_7,                           // 4567        FFFF
    KeyEvent.VK_8, KeyEvent.VK_9, KeyEvent.VK_SEMICOLON, KeyEvent.VK_QUOTE,               // 89:;        FFFF
    KeyEvent.VK_COMMA, KeyEvent.VK_MINUS, KeyEvent.VK_PERIOD, KeyEvent.VK_SLASH,          // <=>?        TTTT
    KeyEvent.VK_OPEN_BRACKET, KeyEvent.VK_A, KeyEvent.VK_B, KeyEvent.VK_C,                // @ABC        FTTT
    KeyEvent.VK_D, KeyEvent.VK_E, KeyEvent.VK_F, KeyEvent.VK_G,                           // DEFG        TTTT
    KeyEvent.VK_H, KeyEvent.VK_I, KeyEvent.VK_J, KeyEvent.VK_K,                           // HIJK        TTTT
    KeyEvent.VK_L, KeyEvent.VK_M, KeyEvent.VK_N, KeyEvent.VK_O,                           // LMNO        TTTT
    KeyEvent.VK_P, KeyEvent.VK_Q, KeyEvent.VK_R, KeyEvent.VK_S,                           // PQRS        TTTT
    KeyEvent.VK_T, KeyEvent.VK_U, KeyEvent.VK_V, KeyEvent.VK_W,                           // TUVW        TTTT
    KeyEvent.VK_X, KeyEvent.VK_Y, KeyEvent.VK_Z, KeyEvent.VK_ALT_GRAPH,                   // XYZ[        TTTF
    KeyEvent.VK_BACK_SLASH, KeyEvent.VK_CLOSE_BRACKET, KeyEvent.VK_EQUALS, KeyEvent.VK_0, // \]^_        FFFT
    KeyEvent.VK_BACK_SLASH, KeyEvent.VK_A, KeyEvent.VK_B, KeyEvent.VK_C,                  // `abc        TFFF
    KeyEvent.VK_D, KeyEvent.VK_E, KeyEvent.VK_F, KeyEvent.VK_G,                           // defg        FFFF
    KeyEvent.VK_H, KeyEvent.VK_I, KeyEvent.VK_J, KeyEvent.VK_K,                           // hijk        FFFF
    KeyEvent.VK_L, KeyEvent.VK_M, KeyEvent.VK_N, KeyEvent.VK_O,                           // lmno        FFFF
    KeyEvent.VK_P, KeyEvent.VK_Q, KeyEvent.VK_R, KeyEvent.VK_S,                           // pqrs        FFFF
    KeyEvent.VK_T, KeyEvent.VK_U, KeyEvent.VK_V, KeyEvent.VK_W,                           // tuvw        FFFF
    KeyEvent.VK_X, KeyEvent.VK_Y, KeyEvent.VK_Z, KeyEvent.VK_ALT_GRAPH,                   // xyz{        FFFT
    KeyEvent.VK_OPEN_BRACKET, KeyEvent.VK_CLOSE_BRACKET                                  // |}          TT
  };
  
  protected static final String ASC_TO_SHIFT =
    "FTTTTTTTTTTTFFFFFFFFFFFFFFFFTTTTFTTTTTTTTTTTTTTTTTTTTTTTTTTFFFFTTFFFFFFFFFFFFFFFFFFFFFFFFFFTTT";
  
  public static int convertCharToVK(char in) {
    int result;
    switch(in) {
      case '\n': shift = false; result = KeyEvent.VK_ENTER; break;
      case '£':  shift = true; result = KeyEvent.VK_EQUALS; break;
      case 'ö': shift = false; result = KeyEvent.VK_NUMPAD4; break;
      case 'ä': shift = false; result = KeyEvent.VK_NUMPAD6; break;
      case 'ü': shift = false; result = KeyEvent.VK_NUMPAD8; break;
      case 'ß': shift = false; result = KeyEvent.VK_NUMPAD2; break;
      case 'Ö': shift = false; result = KeyEvent.VK_NUMPAD5; break;
      case 'Ä': shift = false; result = KeyEvent.VK_NUMPAD0; break;
    default:
      if (in >= ' ' && in <= '}') {
        in -= 32;
        result = ASC_TO_KEY[in];
        shift = ASC_TO_SHIFT.charAt(in) == 'T';
      }
      else {
        shift = false;
        result = KeyEvent.CHAR_UNDEFINED;
      }
  }
  return result;
}

  protected static final String SNA_HEADER                             = "MV - SNA";
  protected static final String DSK_HEADER                             = "MV - CPC";
  protected static final String DSK_HEADER_EXT                         = "EXTENDED";
  protected static final String CDT_HEADER                             = "ZXTAPE";

  protected static final int    CRTC_FLAG_VSYNC_ACTIVE                 = 0x01;
  protected static final int    CRTC_FLAG_HSYNC_ACTIVE                 = 0x02;
  protected static final int    CRTC_FLAG_HDISP_ACTIVE                 = 0x04;
  protected static final int    CRTC_FLAG_VDISP_ACTIVE                 = 0x08;
  protected static final int    CRTC_FLAG_HTOT_REACHED                 = 0x10;
  protected static final int    CRTC_FLAG_VTOT_REACHED                 = 0x20;
  protected static final int    CRTC_FLAG_MAXIMUM_RASTER_COUNT_REACHED = 0x40;

  protected static final int    SNAPSHOT_ID                            = 0x0000;
  protected static final int    VERSION                                = 0x0010;

  protected static final int    AF                                     = 0x0011;
  protected static final int    BC                                     = 0x0013;
  protected static final int    DE                                     = 0x0015;
  protected static final int    HL                                     = 0x0017;
  protected static final int    R                                      = 0x0019;
  protected static final int    I                                      = 0x001a;
  protected static final int    IFF1                                   = 0x001b;
  protected static final int    IFF2                                   = 0x001c;
  protected static final int    IX                                     = 0x001d;
  protected static final int    IY                                     = 0x001f;
  protected static final int    SP                                     = 0x0021;
  protected static final int    PC                                     = 0x0023;
  protected static final int    IM                                     = 0x0025;
  protected static final int    AF1                                    = 0x0026;
  protected static final int    BC1                                    = 0x0028;
  protected static final int    DE1                                    = 0x002a;
  protected static final int    HL1                                    = 0x002c;

  protected static final int    GA_PEN                                 = 0x002e;
  protected static final int    GA_INKS                                = 0x002f;
  protected static final int    GA_ROM                                 = 0x0040;
  protected static final int    GA_RAM                                 = 0x0041;

  protected static final int    CRTC_REG                               = 0x0042;
  protected static final int    CRTC_REGS                              = 0x0043;

  protected static final int    UPPER_ROM                              = 0x0055;

  protected static final int    PPI_A                                  = 0x0056;
  protected static final int    PPI_B                                  = 0x0057;
  protected static final int    PPI_C                                  = 0x0058;
  protected static final int    PPI_CONTROL                            = 0x0059;

  protected static final int    PSG_REG                                = 0x005a;
  protected static final int    PSG_REGS                               = 0x005b;

  protected static final int    MEM_SIZE                               = 0x006b;

  protected static final int    CPC_TYPE                               = 0x006c;

  protected static final int    VER_INT_BLOCK                          = 0x006d;
  protected static final int    VER_MODES                              = 0x006e;

  protected static final int    HEADER_SIZE                            = 0x0100;

  public void loadFile(int type, String name) throws Exception {
    byte[] data = getFile(name);
    if (SNA_HEADER.equals(new String(data, 0, SNA_HEADER.length()).toUpperCase())){
        SNA_Load(name);
        System.out.println("Loading MV - SNA snapshot file...");
    }
    else
    if (DSK_HEADER.equals(new String(data, 0, DSK_HEADER.length()).toUpperCase())){
        DSK_Load(name);
        System.out.println("Loading MV - CPCEMU DSK file...");
    }
    else
    if (DSK_HEADER_EXT.equals(new String(data, 0, DSK_HEADER_EXT.length()).toUpperCase())){
        DSK_Load(name);
        System.out.println("Loading EXTENDED CPC DSK file...");
    }
    else
    if (CDT_HEADER.equals(new String(data, 0, CDT_HEADER.length()).toUpperCase())){
        CDT_Load(name);
        System.out.println("Loading ZXTAPE file...");
    }
    else
        System.out.println("Unrecognized file format");
    psg.getSoundPlayer().resync();
    fdc.initialise();
    }
  
  public void CDT_Load(String name) throws Exception {
    byte[] data = getFile(name);
      Switches.booter = 0;
      //CPCTapeImage image = new CPCTapeImage(name, data);
      //
      // TODO: CDT support
      //
       System.out.println("No CDT/TZX support now...");
  }
  
  public void DSK_Load(String name) throws Exception {
    byte[] data = getFile(name);
      Switches.booter = 0;
      CPCDiscImage image = new CPCDiscImage(name, data);
      int drive = getCurrentDrive();
      if (drive == 0)          
      dskImageA = image;
      if (drive == 1)          
      dskImageB = image;
      if (drive == 2)          
      dskImageC = image;
      if (drive == 3)          
      dskImageD = image;
      floppies[drive].setSides(image.getNumberOfSides());
      int heads = image.getNumberOfSides() == 1 ? Drive.HEAD_0 : Drive.BOTH_HEADS;
      fdc.setDrive(drive, floppies[drive]);
      fdc.getDrive(drive).setDisc(heads, image);
      if (getCurrentDrive() == 0)
        Switches.loaddrivea = name;
      if (getCurrentDrive() == 1)
        Switches.loaddriveb = name;
      if (getCurrentDrive() == 2)
        Switches.loaddrivec = name;
      if (getCurrentDrive() == 3)
        Switches.loaddrived = name;
      
      if (Switches.floppysound){
        UPD765A.insertSound.play();
        UPD765A.motorSound.play();
      }
  }
  
  public void SNA_Load(String name){
    byte[] data = getFile(name);
      Switches.booter = 1;
      z80.setAF(getWord(data, AF));
      z80.setBC(getWord(data, BC));
      z80.setDE(getWord(data, DE));
      z80.setHL(getWord(data, HL));
      z80.setR(data[R]);
      z80.setI(data[I]);
      z80.setIFF1(data[IFF1] != 0);
      z80.setIFF2(data[IFF2] != 0);
      z80.setIX(getWord(data, IX));
      z80.setIY(getWord(data, IY));
      z80.setSP(getWord(data, SP));
      z80.setPC(getWord(data, PC));
      z80.setIM(data[IM]);
      z80.setAF1(getWord(data, AF1));
      z80.setBC1(getWord(data, BC1));
      z80.setDE1(getWord(data, DE1));
      z80.setHL1(getWord(data, HL1));

      gateArray.setSelectedInk(data[GA_PEN]);
      System.out.print("INK READ: ");
      for (int i = 0; i < 0x11; i++){
        gateArray.setInk(i, data[GA_INKS + i]);
        System.out.print(data[GA_INKS + i]);
        if (i>=0x10)
            System.out.println("");
        else
            System.out.print(",");
      }
      
      gateArray.setModeAndROMEnable(memory, data[GA_ROM]);
      memory.setRAMBank(data[GA_RAM]);

      crtc.setSelectedRegister(data[CRTC_REG]);
      for (int i = 0; i < 18; i++)
        crtc.setRegister(i, data[CRTC_REGS + i]);

      ppi.setControl(data[PPI_CONTROL] & 0xff | 0x80);
      ppi.setOutputValue(PPI8255.PORT_A, data[PPI_A] & 0xff);
      ppi.setOutputValue(PPI8255.PORT_B, data[PPI_B] & 0xff);
      int portC = data[PPI_C] & 0xff;
      ppi.setOutputValue(PPI8255.PORT_C, portC);

      psg.setBDIR_BC2_BC1(PSG_VALUES[portC >> 6], ppi.readOutput(PPI8255.PORT_A));

      psg.setSelectedRegister(data[PSG_REG]);
      for (int i = 0; i < 14; i++)
        psg.setRegister(i, data[PSG_REGS + i] & 0xff);

      int memSize = (data[MEM_SIZE] & 0xff) * 1024;
      memSize = Math.min(memSize, memory.getRAMType() == CPCMemory.TYPE_512K ? 0x10000 : 0x20000);
      byte[] mem = memory.getMemory();
      System.arraycopy(data, 0x100, mem, 0, memSize);

  }
  
  public void displayLostFocus() {
        keyboardb.reset();
        keyboarda.reset();
        keyboardn.reset();
  }
  
  public Drive[] getFloppyDrives() {
    return floppies;
  }
  
}